/*
 *  IonCfg - Lightweight multiuse configuration system
 *  Copyright (C) 2011-2014  Alicia Cathryn Coder
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ioncfg.h"

const char* cvalue(struct configgroup* cfg, const char* property)
{
  unsigned int i;
  for(i=0; i<cfg->valuecount; ++i)
  {
    if(!strcmp(cfg->values[i].name, property)){return cfg->values[i].value;}
  }
  return 0;
}

struct configgroup* cgroup(struct configgroup* cfg, const char* type, const char* name)
{
  unsigned int count;
  struct configgroup* grouplist=cgrouplist(cfg, type, &count);
  if(!grouplist){return 0;}
  unsigned int i;
  for(i=0; i<count; ++i)
  {
    if(!strcmp(name, grouplist[i].name)){return &grouplist[i];}
  }
  return 0;
}

struct configgroup* cgrouplist(struct configgroup* cfg, const char* type, unsigned int* count)
{
  unsigned int i;
  for(i=0; i<cfg->grouptypecount; ++i)
  {
    if(!strcmp(cfg->grouptypes[i].type, type))
    {
      (*count)=cfg->grouptypes[i].groupcount;
      return cfg->grouptypes[i].groups;
    }
  }
  return 0;
}

void readconfig(FILE* file, struct configgroup* group)
{
  unsigned int valuelen=8;
  group->valuecount=0;
  group->values=malloc(sizeof(struct configvalue)*valuelen);
  unsigned int grouptypelen=8;
  group->grouptypecount=0;
  group->grouptypes=malloc(sizeof(struct configgrouptype)*grouptypelen);
  char buf[1024]; // Bad limit, maybe
  while(fgets(buf, 1024, file))
  {
    unsigned int end=strlen(buf);
    while(end>0&&(buf[end-1]=='\r'||buf[end-1]=='\n'))
    {
      --end;
    }
    buf[end]=0;
    char* line=buf;
    while(line[0]==' '||line[0]=='\t'){++line;}
    if(line[0]=='#'||!line[0]){continue;} // Comment! (or empty line)
    if(line[0]=='}'){break;} // End of group
    if(strchr(line, '='))
    {
      char* eq=strchr(line, '=');
      eq[0]=0;
      if(group->valuecount>=valuelen) // Need to expand
      {
        valuelen+=8;
        group->values=realloc(group->values, sizeof(struct configvalue)*valuelen);
      }
      group->values[group->valuecount].name=strdup(line);
      group->values[group->valuecount].value=strdup(eq+1);
      ++group->valuecount;
    }
    else if(strchr(line, '{'))
    {
      // TODO, for lines that are neither comments, ends of group or values, save them in case there's a '{' on the next line
      // Separate type and name (if any)
      strchr(line, '{')[0]=0;
      end=strlen(line);
      while(end>0&&(line[end-1]==' '||line[end-1]=='\t'))
      {
        --end;
      }
      line[end]=0;
      char* name=strchr(line, ' ');
      if(name)
      {
        name[0]=0;
        ++name;
      }else{
        name="";
      }
      // Check if the type exists first..
      unsigned int type;
      for(type=0; type<group->grouptypecount; ++type)
      {
        if(!strcmp(line, group->grouptypes[type].type)){break;}
      }
      if(type>=group->grouptypecount) // New grouptype
      {
        if(type>=grouptypelen) // Need to expand
        {
          grouptypelen+=8;
          group->grouptypes=realloc(group->grouptypes, sizeof(struct configgrouptype)*grouptypelen);
        }
        group->grouptypes[type].type=strdup(line);
        group->grouptypes[type].groupcount=0;
        group->grouptypes[type].grouplen=8;
        group->grouptypes[type].groups=malloc(sizeof(struct configgroup)*group->grouptypes[type].grouplen);
        ++group->grouptypecount;
      }
      if(group->grouptypes[type].groupcount>=group->grouptypes[type].grouplen) // Need to expand
      {
        group->grouptypes[type].grouplen+=8;
        group->grouptypes[type].groups=realloc(group->grouptypes[type].groups, sizeof(struct configgroup)*group->grouptypes[type].grouplen);
      }
      unsigned int id=group->grouptypes[type].groupcount;
      group->grouptypes[type].groups[id].name=strdup(name);
      ++group->grouptypes[type].groupcount;
      readconfig(file, &group->grouptypes[type].groups[id]);
    }
  }
  // Reallocate to only use what we need
  if(group->valuecount<valuelen)
  {
    if(group->valuecount>0)
    {
      group->values=realloc(group->values, sizeof(struct configvalue)*group->valuecount);
    }else{
      free(group->values);
      group->values=0;
    }
  }
  if(group->grouptypecount<grouptypelen)
  {
    if(group->grouptypecount>0)
    {
      group->grouptypes=realloc(group->grouptypes, sizeof(struct configgrouptype)*group->grouptypecount);
    }else{
      free(group->grouptypes);
      group->grouptypes=0;
    }
  }
}

struct configgroup* loadconfig(const char* filename)
{
  FILE* file=fopen(filename, "r");
  if(!file){return 0;}
  struct configgroup* group=malloc(sizeof(struct configgroup));
  group->name=0;
  readconfig(file, group);
  fclose(file);
  return group;
}

void freecgroup(struct configgroup* cfg)
{
  unsigned int i;
  for(i=0; i<cfg->grouptypecount; ++i)
  {
    free((void*)cfg->grouptypes[i].type);
    unsigned int i2;
    for(i2=0; i2<cfg->grouptypes[i].groupcount; ++i2)
    {
      freecgroup(&cfg->grouptypes[i].groups[i2]);
    }
    free(cfg->grouptypes[i].groups);
  }
  free(cfg->grouptypes);
  for(i=0; i<cfg->valuecount; ++i)
  {
    free((void*)cfg->values[i].name);
    free((void*)cfg->values[i].value);
  }
  free(cfg->values);
  free((void*)cfg->name);
}

void freeconfig(struct configgroup* cfg)
{
  freecgroup(cfg);
  free(cfg);
}
